class AYFile
	def self.read(filename)
		self.new.read(filename)
	end
	
	attr_reader :file_version, :player_version
	attr_accessor :first_song, :author, :misc, :songs
	
	def initialize(opts = {})
		@file_version = opts[:file_version] || 1
		@player_version = opts[:player_version] || 0
		@first_song = opts[:first_song] || 0
		@author = opts[:author] || ''
		@misc = opts[:misc] || ''
		@songs = opts[:songs] || []
	end
	
	def read(filename)
		data = File.read(filename)

		(
			signature, # offset 0
			@file_version, # 8
			@player_version, # 9
			p_special_player, # 10
			p_author, # 12
			p_misc, # 14
			num_of_songs, # 16
			first_song, # 17
			p_songs_structure, # 18
		) = data.unpack('a8CCnnnCCn')
		raise "File is not a valid AY file" if signature != 'ZXAYEMUL'
		raise "File has a special player - not supported" if p_special_player != 0
		
		@author, = data[(p_author + 12)..-1].unpack('Z*')
		@misc, = data[(p_misc + 14)..-1].unpack('Z*')
		
		num_of_songs += 1
		
		@songs = []
		offset = p_songs_structure + 18
		num_of_songs.times do
			p_song_name, p_song_data = data[offset, 4].unpack('nn')
			song_name, = data[(offset + p_song_name)..-1].unpack('Z*')
			song_data_offset = offset + 2 + p_song_data
			song_data = data[song_data_offset, 14]
			(
				a_chan, b_chan, c_chan, noise, # offset 0
				song_length, # 4
				fade_length, # 6
				hi_reg, # 8
				lo_reg, # 9
				p_points, # 10
				p_addresses, # 12
			) = song_data.unpack('CCCCnnCCnn')
			
			points_data_offset = song_data_offset + p_points + 10
			blocks_data_offset = song_data_offset + p_addresses + 12
			
			points_data = data[points_data_offset, 6]
			stack, init, inter = points_data.unpack('nnn')
			
			blocks = []
			
			until data[blocks_data_offset] == 0 and data[blocks_data_offset + 1] == 0
				address, length, block_offset = data[blocks_data_offset, 6].unpack('nnn')
				if address + length > 0x10000
					length = 0x10000 - address
				end
				block = data[blocks_data_offset + 4 + block_offset, length]
				blocks << MemoryBlock.new(address, block)
				blocks_data_offset += 6
			end
			
			@songs << Song.new(
				:name => song_name,
				:a_chan => a_chan, :b_chan => b_chan, :c_chan => c_chan,
				:noise => noise,
				:song_length => song_length, :fade_length => fade_length,
				:hi_reg => hi_reg, :lo_reg => lo_reg,
				:stack => stack, :init => init, :inter => inter,
				:blocks => blocks)

			offset += 4
		end
		
		self
	end
	
	def write(filename)
		# determine global offsets for everything
		current_offset = 20 # length of file header
		songs_structure_offset = current_offset
		song_structure_offsets = []
		songs.each_index do |i|
			song_structure_offsets[i] = current_offset
			current_offset += 4
		end

		song_data_offsets = []
		songs.each_index do |i|
			song_data_offsets[i] = current_offset
			current_offset += 14
		end

		points_offsets = []
		songs.each_index do |i|
			points_offsets[i] = current_offset
			current_offset += 6
		end
		
		blocks_data_offsets = []
		songs.each_with_index do |song, i|
			blocks_data_offsets[i] = current_offset
			current_offset += song.blocks.size * 6 + 2
		end
		
		author_offset = current_offset
		current_offset += @author.length + 1
		misc_offset = current_offset
		current_offset += @misc.length + 1
		
		song_names_offsets = []
		songs.each_with_index do |song, i|
			song_names_offsets[i] = current_offset
			current_offset += song.name.length + 1
		end
		
		memory_blocks_offsets = {}
		memory_data = ''
		@songs.each do |song|
			song.blocks.each do |block|
				next if memory_blocks_offsets[block.data]
				memory_blocks_offsets[block.data] = current_offset
				memory_data << block.data
				current_offset += block.data.size
			end
		end
		
		File.open(filename, 'wb') do |f|
			# write file header
			f.write [
				'ZXAYEMUL', @file_version, @player_version, 0,
				author_offset - 12, misc_offset - 14,
				@songs.size - 1, @first_song, songs_structure_offset - 18
			].pack('A8ccnnnccn')

			current_offset = 20
			# write song structures
			@songs.each_index do |i|
				f.write [
					song_names_offsets[i] - current_offset,
					song_data_offsets[i] - current_offset - 2
				].pack('nn')
				current_offset += 4
			end

			# write song data
			@songs.each_with_index do |song, i|
				f.write [
					song.a_chan, song.b_chan, song.c_chan, song.noise,
					song.song_length, song.fade_length,
					song.hi_reg, song.lo_reg,
					points_offsets[i] - current_offset - 10,
					blocks_data_offsets[i] - current_offset - 12
				].pack('ccccnnccnn')
				current_offset += 14
			end
			
			# write points data
			@songs.each do |song|
				f.write [
					song.stack, song.init, song.inter
				].pack('nnn')
				current_offset += 6
			end
			
			# write blocks data
			@songs.each do |song|
				song.blocks.each do |block|
					f.write [
						block.address,
						block.data.length,
						memory_blocks_offsets[block.data] - current_offset - 4
					].pack('nnn')
					current_offset += 6
				end
				f.write [0].pack('n')
				current_offset += 2
			end
			
			# write author / misc strings (null-terminated)
			f.write [@author, @misc].pack('Z*Z*')
			
			# write song names
			@songs.each do |song|
				f.write [song.name].pack('Z*')
			end
			
			# write memory blocks
			f.write memory_data
		end
	end
	
	class Song
		attr_accessor :name, :a_chan, :b_chan, :c_chan, :noise,
			:song_length, :fade_length, :hi_reg, :lo_reg,
			:stack, :init, :inter, :blocks
		def initialize(opts = {})
			@name = opts[:name] || ''
			@a_chan = opts[:a_chan] || 0
			@b_chan = opts[:b_chan] || 1
			@c_chan = opts[:c_chan] || 2
			@noise = opts[:noise] || 3
			@song_length = opts[:song_length] || 0
			@fade_length = opts[:fade_length] || 0
			@hi_reg = opts[:hi_reg] || 1
			@lo_reg = opts[:lo_reg] || 0
			@stack = opts[:stack] || 0
			@init = opts[:init] || 0
			@inter = opts[:inter] || 0
			@blocks = opts[:blocks] || []
		end
	end

	MemoryBlock = Struct.new(:address, :data)

end
